from typing import Iterable

from typeguard import typechecked

from .choc_expr import ChocExpr
from .col import Col
from ..utils import quote_expr


@typechecked
class ColWhen(ChocExpr):
    """ 'Case When' expression  """
    def __init__(
            self,
            col: Col | str,
            expected_vals: Iterable[str | int | float],
            returned_vals: Iterable[str | int | float],
            else_returned_val: str | int | float | None = None,
            compact: bool = False
    ):
        if len(expected_vals) != len(returned_vals):
            raise ValueError("'conditions' and 'returned_vals' must have the same length")

        if len(returned_vals) == 0:
            raise ValueError("CaseWhen must not be empty")

        self._col = (Col(col) if type(col) is str else col).remove_alias()
        self._when_then_items = [
            f"WHEN {quote_expr(expected_val)} THEN {quote_expr(returned_val)}" for expected_val, returned_val in zip(expected_vals, returned_vals)
        ]
        self._else = f"ELSE {quote_expr(else_returned_val)}" if else_returned_val else ""
        super().__init__("CASE {_col.full_name~}{$(_when_then_items)~}{_else~}END", compact=compact, list_join_sep="\n")



from typing import Iterable, Self, Tuple

from typeguard import typechecked

from .delete_from import DeleteFrom
from ..query_mode import QueryMode
from .join import Join
from .order_by import OrderBy
from .group_by import GroupBy
from .having import Having
from .where import Where
from .select_from import SelectFrom
from .limit import Limit
from ..agg_function import AggFunction
from .choc_expr import ChocExpr
from ..join_type import JoinType
from .condition import Condition
from .col import Col
from .table import Table
from .update_set import UpdateSet


@typechecked
class Query(ChocExpr):
    """ Handler to generate a SQL query """
    def __init__(
            self,
            query_mode: QueryMode = QueryMode.Select,
            compact: bool = True,
            limit: int | None = None,
            table: str | Table | None = None,
            unique: bool = False,
            joins: Iterable[Tuple[str | Table, Condition | str | Iterable[str]] | Tuple[str | Table, Condition | str | Iterable[str], JoinType | None]] | None = None,
            cols: Iterable[str | Col] | None = None,
            groups: Iterable[str] | None = None,
            filters: Iterable[Condition] | None = None,
            assignations: Iterable[Condition] | None = None
    ) -> None:
        if cols is None:
            cols = []
        if joins is None:
            joins = []
        if groups is None:
            groups = []
        if filters is None:
            filters = []
        if assignations is None:
            assignations = []
        if query_mode == QueryMode.Select and assignations:
            raise ValueError("You cannot have update assignations in read query mode")
        elif query_mode == QueryMode.Update and cols:
            raise ValueError("You cannot have update cols in update query mode")
        self._query_mode = query_mode
        self._select_from = SelectFrom(table=table, unique=unique, compact=False)
        self._order_by = OrderBy(select=self._select_from.select, compact=False)
        self._group_by = GroupBy(compact=False)
        self._limit = Limit(length=limit, compact=False)
        self._where = Where(compact=False)
        self._having = Having(compact=False)
        self._joins = []
        self._compact = compact
        self._update_set = UpdateSet(table=table, assignations=assignations, compact=compact)
        self._delete_from = DeleteFrom(table=table, compact=compact)
        if joins:
            self.join_many(*joins)
        if cols and query_mode == QueryMode.Select:
            self.select(*cols)
        if groups:
            self.group_by(*groups)
        for filter in filters:
            self.filter(filter)
        super().__init__(
            "@{update_mode}:{_update_set~}:;@{delete_mode}:{_delete_from~}:;@{read_mode}:{_select_from~}{$(_joins)~}:;{_where~}@{read_mode}:{_group_by~}{_having~}{_order_by~}{_limit~}:;",
            list_join_sep="\n",
            compact=compact
        )

    @property
    def read_mode(self) -> bool:
        return self._query_mode == QueryMode.Select

    @property
    def update_mode(self) -> bool:
        return self._query_mode == QueryMode.Update

    @property
    def delete_mode(self) -> bool:
        return self._query_mode == QueryMode.Delete

    def table(self, val: str | Table | None) -> Self:
        """ Set the table name """
        self._select_from.from_expr.table = val
        return self

    def select(self, *vals: str | Col) -> Self:
        """ Set the selected cols """
        self._query_mode = QueryMode.Select
        self._select_from.select.cols = vals
        return self

    def update(self, *assignations: Condition) -> Self:
        """ Update """
        self._query_mode = QueryMode.Update
        self._update_set._set.assignations = assignations
        return self

    def delete(self) -> Self:
        """ Delete """
        self._query_mode = QueryMode.Delete
        return self

    def distinct(self) -> Self:
        """ Filter the rows to remove duplicates (by selected columns)"""
        self._select_from.select.unique = True
        return self

    def head(self, length: int = 1) -> Self:
        """ Filter on the first N rows """
        self._limit = length
        return self

    def filter(self, condition: Condition) -> Self:
        """ Filter the rows according to the given condition """
        if any(x in condition.build() for x in set(e.value for e in AggFunction)):
            self._having.condition = condition
        else:
            self._where.condition = condition
        return self

    def group_by(self, *cols_names: str) -> Self:
        """ Group the rows of the specified columns """
        self._group_by.cols = cols_names
        return self

    def join(self, table: str | Table, condition: Condition | str | Iterable[str], join_type: JoinType | None = JoinType.Inner) -> Self:
        self._joins.append(Join(table=table if isinstance(table, Table) else Table(name=table), condition=condition, join_type=join_type, compact=self._compact))
        return self

    def join_many(self, *joins_params: Tuple[str | Table, Condition | str | Iterable[str]] | Tuple[str | Table, Condition | str | Iterable[str], JoinType | None]) -> Self:
        for join_params in joins_params:
            self.join(*join_params)
        return self



import re
from typing import Any, List, Tuple

from .expr import Expr


class ChocExprAttr(Expr):

    def __init__(
            self,
            obj: Any,
            expr: str,
            list_join_sep: str = ", "
    ) -> None:
        self._expr = expr
        self._obj = obj
        self._list_join_sep = list_join_sep
        self._endline = False

    @property
    def list_join_sep(self) -> str:
        return self._list_join_sep

    @list_join_sep.setter
    def list_join_sep(self, value: str) -> None:
        self._list_join_sep = value

    def _get_nested_value(self, obj: Any, keys: List[str]) -> Any:
        for key in keys:
            obj = getattr(obj, key)
        return obj

    def _get_nested_value_from_expr(self, obj: Any, expr: str) -> Any:
        return self._get_nested_value(obj, expr.split("."))

    def _extract_list_expr(self, expr: str) -> Tuple[None | str, str]:
        if m := re.search(r"\$\(([\w.,~]+)\)", expr):
            if not expr.startswith("$("):
                raise ValueError("$ statement must be placed at the begining")
            return m[0][2:-1], expr[len(m[0]) + 1:]
        return None, expr

    def _check_endline(self) -> bool:
        if "~" in self._expr:
            if self._expr.endswith("~"):
                self._endline = True
                self._expr = self._expr[:-1]
            else:
                raise ValueError("The '~' must be placed at end")

    def _get_value(self) -> Any:
        lst_expr, sub_expr = self._extract_list_expr(self._expr)
        val = [str(self._get_nested_value_from_expr(x, sub_expr) if sub_expr else x) for x in self._get_nested_value_from_expr(self._obj, lst_expr)] if lst_expr else \
            str(self._get_nested_value_from_expr(self._obj, self._expr))
        return val

    def build(self) -> str:
        self._check_endline()
        val = self._get_value()
        expr = self.list_join_sep.join(x for x in val if x) if isinstance(val, List) else val
        if self._endline and expr != "":
            expr += "\n"
        return expr



from abc import ABC, abstractmethod


class Expr(ABC):
    """ Expression """

    def __str__(self) -> str:
        return self.build()

    def __expr__(self) -> str:
        return self.build()

    @property
    def buildable(self) -> bool:
        return True

    @abstractmethod
    def build(self) -> str:
        pass



import re
from typing import Any

from typeguard import typechecked

from ..utils import to_bool
from .choc_expr_attr import ChocExprAttr
from .expr import Expr


@typechecked
class ChocExpr(Expr):
    """ Choc Expr """

    def __init__(
        self,
        choc_expr: str | None = None,
        list_join_sep: str = ", ",
        compact: bool = True
    ) -> None:
        self._expr = choc_expr or ""
        self.list_join_sep = list_join_sep
        self.compact = compact

    @property
    def list_join_sep(self) -> str:
        return self._list_join_sep

    @list_join_sep.setter
    def list_join_sep(self, value: str) -> None:
        self._list_join_sep = value

    @property
    def compact(self) -> bool:
        return self._compact

    @compact.setter
    def compact(self, value: bool) -> None:
        self._compact = value

    @staticmethod
    def eval_conditions(obj: Any, expr: str) -> str:
        regex = r"(@{([A-Za-z_.$()]+)}:([{}A-Za-z_.$()\s*~]*):([{}A-Za-z_.$()\s*~]*);)"
        matches = re.findall(regex, expr)
        to_replace = {cond: cond_attr_if_true if to_bool(ChocExprAttr(obj, cond_attr).build()) else cond_attr_if_false for cond, cond_attr, cond_attr_if_true, cond_attr_if_false in matches}
        for cond, replacement in to_replace.items():
            expr = expr.replace(cond, replacement)
        return expr

    def eval_attributes(self, expr: str) -> str:
        if not self.buildable:
            return ""
        regex = r"{([A-Za-z_.$()~]+)}"
        to_replace = [key for key in re.findall(regex, expr)]
        for key in to_replace:
            expr = expr.replace(f"{{{key}}}", ChocExprAttr(obj=self, expr=key, list_join_sep=self.list_join_sep).build())
        return expr

    def build(self) -> str:
        expr = self.eval_attributes(ChocExpr.eval_conditions(self, self._expr))
        if self.compact:
            expr = expr.replace("\n", " ")
            expr = expr.strip()
        return expr
